/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_common.h"
#include "nest_proxy.h"

#include "nest_concurrency_manager.h"
#include "nest_collection_manager.h"
#include "nest_storage_manager.h"
#include "nest_client_pool.h"
#include "nest_protocol_manager.h"
#include "nest_advertiser.h"
#include "nest_protocol_advertise.h"
#include "nest_config.h"
#include "nest_hash_table.h"
#include "nest_filecache_model.h"
#include "nest_user.h"
#include "nest_version.h"

#include <sys/utsname.h>
#include <rpc/rpc.h>

    // time between any regular maintenance (in millisecs)
#define MAINTENANCE_INTERVAL 100000

    // handlers for the various services provided to clients
nestRequest* ReceiveRequest( nestClient *client, fd_set *ready );
int  HandleRequests( nestClientPool *requests, fd_set *readfds );
int  HandleRequest( nestClient *client );
NestReplyStatus NewClient( nestClient *newclient );
bool RemoveClient ( nestClient *client, fd_set *reads );
int  SynchronousRequestHandler( nestClient *client, nestRequest * );
NestReplyStatus SendNestAd( nestClient *, nestRequest * );
NestReplyStatus openStorageConnections( nestProtocol **storage,
                                        nestProtocol **otherStorage,
                                        nestClient   *client);
NestReplyStatus SendDirectoryList( nestClient *, nestRequest * );
NestReplyStatus SendPwd( nestClient *, nestRequest * );
NestReplyStatus MakeLink( nestClient *, nestRequest * );
NestReplyStatus CollectionOp( nestClient *, nestRequest * );
NestReplyStatus DirOp( nestClient *, nestRequest * );
NestReplyStatus SendVersion( nestClient *, nestRequest * );
int             SendBandwidthStatistics( nestClient *, nestRequest * );
int             TransferFile( nestClient *, nestRequest *,
							  bool already_trans = false );
NestReplyStatus CopyFile ( nestClient *, nestRequest * );
NestReplyStatus DeleteFile ( nestClient *, nestRequest * );
NestReplyStatus RenameFile ( nestClient *, nestRequest * );
NestReplyStatus SendFileinfo( nestClient *, nestRequest * );
NestReplyStatus SetFileAttr( nestClient *, nestRequest * );
NestReplyStatus SendFilesize( nestClient *, nestRequest * );
int             NestSync( nestClient *, nestRequest * );
NestReplyStatus QueryProxys ( nestClient *, nestRequest * );
NestReplyStatus RegisterProxyClient ( nestClient *, nestRequest * );
NestReplyStatus RemoveProxyClient ( nestClient * );
NestReplyStatus FormatDisk ( nestClient *, nestRequest * );
NestReplyStatus AuthenticateUser( nestClient *, nestRequest * );
NestReplyStatus AddUser( nestClient *, nestRequest * );
NestReplyStatus AddGroup( nestClient *, nestRequest * );
NestReplyStatus DeleteUser( nestClient *, nestRequest * );
NestReplyStatus DeleteGroup( nestClient *, nestRequest * );
NestReplyStatus AddUserToGroup( nestClient *, nestRequest * );
NestReplyStatus DeleteUserFromGroup( nestClient *, nestRequest * );
NestReplyStatus ReserveUserSpace( nestClient *, nestRequest * );
NestReplyStatus ReserveGroupSpace( nestClient *, nestRequest * );
NestReplyStatus ListUserLots( nestClient *, nestRequest * );
NestReplyStatus ListGroupLots( nestClient *, nestRequest * );
NestReplyStatus GetLotStat( nestClient *, nestRequest * );
NestReplyStatus UpdateLot( nestClient *, nestRequest * );
NestReplyStatus ChangeLotType( nestClient *, nestRequest * );
NestReplyStatus MoveFileLot( nestClient *, nestRequest * );
NestReplyStatus AddUserToLot( nestClient *, nestRequest * );
NestReplyStatus DeleteUserFromLot( nestClient *, nestRequest * );
NestReplyStatus AttachLotToFile( nestClient *, nestRequest * );
NestReplyStatus DetachLotFromFile( nestClient *, nestRequest * );
NestReplyStatus SendPhysName( nestClient *, nestRequest * );
NestReplyStatus SendUserAcl( nestClient *, nestRequest * );
NestReplyStatus SetUserAcl( nestClient *, nestRequest * );
NestReplyStatus TerminateLot( nestClient *, nestRequest * );
NestReplyStatus DebugRequest( nestClient *, nestRequest * );
NestReplyStatus TerminateServer( nestClient *, nestRequest * );

void taskDone( nestClient *client, NestReplyStatus status );


NestReplyStatus GetFilehandle( nestClient *client, nestRequest * );
NestReplyStatus SendAvailableSpace( nestClient *, nestRequest * );

    // other misc helper functions
NestReplyStatus RegularMaintenance( );
struct timeval* setMaintenanceTimeout( struct timeval *timeout );
int  getMaxfd( );
void installSignalHandlers( );
void sigChild( int signal, siginfo_t *info, void *context );
int  main_init( int argc, const char *argv[] );
NestReplyStatus WriteAd( );
void debug_init( bool Daemonize );
void daemon_init( const char *pname, int facility );

    // timer to do regular maintenance
static struct timeval last_maintenance;
static int maintenance_interval = MAINTENANCE_INTERVAL;

    // client connection states
enum {  CLOSE_CONNECTION,       // client is no longer in the building
        ACTIVE_CONNECTION,      // client is being handled by concurrency mgr
        PASSIVE_CONNECTION      // client should be monitored by server
};

    // for a main nest to know which proxys there are
    // this hash table is an stl list for each client registering proxies
static struct nest_hash_table *registered_proxys = NULL;

    // acceptable authentication models
static int allowedAuthentication = NEST_AUTH_GSS;

    // collection of currently connected clients
static nestClientPool *clients = NULL;

    // collection of server initiated tasks
static nestClientPool *serverTasks = NULL;

    // the pid of the main server thread
    // used to make sure that only main server thread performs protected duties
extern int admin_pid;

    // used to make a nest appear remote
unsigned long latency = 0;

    // needs to be global now for taskDone
fd_set writefds, readfds, r_readyfds, w_readyfds;

	// Storage manager, etc
static nestStorageManager		*storage_manager = NULL;
static nestIdentityManager		*identity_manager = NULL;
static nestFileCatalog			*file_catalog = NULL;
static nestLotManager			*lot_manager = NULL;
static nestCollectionManager	*collection_manager = NULL;
static nestAdvertiser			*advertiser = NULL;
static nestConcurrencyManager	*concurrency_manager = NULL;
static nestProtocolManager		*protocol_manager = NULL;

static const char* NestVersionString =
"$NestVersion: " NEST_VERSION " " __DATE__ " $";

const char *getNestVersionStr( void )
{
	return NestVersionString;
}

int
main( int argc, const char *argv[] )
{

    main_init( argc, argv );

        // enter the main server loop after doing first RegularMaintenance
        // first handle any messages from our dispatched worker processes
        // then check for any new incoming client requests
    RegularMaintenance();
    nestClientPool *requests = new nestClientPool( storage_manager );
    nest_printf( "Initializing done.  Ready to service requests.\n" );
    while( true ) {

            // merge rpc readlist with our own, then enter select loop
        r_readyfds = svc_fdset;
        NestMergeFdsets( &r_readyfds, &readfds, getMaxfd() + 1 );
        nest_debug( NEST_DEBUG_FDSETS, "NestRead set ( %x ): ", &readfds );
        NestFd_print( NEST_DEBUG_FDSETS, &readfds, getMaxfd() + 1 );
        w_readyfds = writefds;
        nest_debug( NEST_DEBUG_FDSETS,
                "maxfd is %d, entering select\n", (getMaxfd()+1) );
        int readies = NestSelect(getMaxfd()+1, &r_readyfds,&w_readyfds, NULL,
                            NestNextTimeout( maintenance_interval, 
                                &last_maintenance ) );
        if ( readies <= 0 ) {
            if ( readies == 0 ) {
                    // TODO need slow timeouts for regular maintenance
                    // and fast timeouts for scheduling
                RegularMaintenance();
            } else nest_warn( __FILE__, __FUNCTION__, 
                    "select failed: %s\n", strerror( errno ) );
            continue;
        }
            // appear remote if specified
            // half of latency should be used because clients will be expected
            // to arrive during some previous sleep?
        if ( latency != 0 ) {
            NestNanosleep( latency * 1000 * 1000 );
        }
        nest_debug( NEST_DEBUG_FDSETS, "Woke from select with %d readies.\n",
                readies );

            // then accept any new clients - listen on each protocol socket
        for( int protocol_socket  = protocol_manager->firstProtocolSocket();
             protocol_socket != protocol_manager->lastProtocolSocket();
             protocol_socket  = protocol_manager->nextProtocolSocket() )
        {
                // ignore any sockets that were unable to listen correctly
            if ( protocol_socket <= 0 ) continue;

                // some protocols may have multiple clients to give . . .
                // some protocols may have none and just use the socket as
                // a way of waking up the main thread (i.e. gftp)
            nestClient *newclient;
            int accepts = 0;
            while ( ( newclient = protocol_manager->accept(
						  protocol_socket, &r_readyfds, accepts ))
					!= NULL )
            {

                accepts++;

                // clear the socket since ftp and gftp share one
                // I don't really understand this anymore . . .
                //NestFd_clear( protocol_socket, &r_readyfds );

                    // add new connection state to appropriate structures
                NewClient( newclient );

            }
        }

            // then queue all requests from existing clients
        nestClient *client;
        nestRequest *req;
        CLIENT_ITERATOR itr = clients->begin();
        int size;
        if ( ( size = requests->clear() ) != 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
					   "Requests list was not previously cleared. Size %d.\n",
					   size );
        }
        while( itr != clients->end( ) ) {
            client = *itr++;
            nest_debug( NEST_DEBUG_CLIENT, "Checking client %d (%s)\n",
						client->getId(), client->getUserIdentity() );

                // pass the r_readyfds to each client, if they are ready
                // they can create a request object
            req = ReceiveRequest( client, &r_readyfds );
            if ( req != NULL ) {
                requests->addClient( client );
                nest_debug( NEST_DEBUG_CLIENT,
							"Client '%s' (%d) request: \"%s\"\n",
							client->getUserIdentity(),
							client->getId(),
							req->toString() );
            }
        } 

            // now iterate through new requests and handle them here
        HandleRequests( requests, &readfds );

            // work on any concurrent transfers last
            // this is important to be last to allow anticipatory scheduling
        concurrency_manager->makeProgress( );

    } // end of main service loop

        // we should never get here
    nest_die( "Big trouble in Little China.\n" );
    return ERROR_EXIT;
}

NestReplyStatus
NewClient( nestClient *newclient )
{
    int size = clients->addClient( newclient );
    nest_debug( NEST_DEBUG_CLIENT,
            "added a new client, client count is %d\n", size );

    if ( newclient->hasSocket() ) {
        nestProtocol *connection = newclient->getConnection();
            // may be an optimization to add to ready_fds
        NestFd_set( connection->getSocket(),&readfds);
            // I don't think this is an optimization
        //NestFd_set( connection->getSocket(), &r_readyfds );
    }
    return NEST_SUCCESS;
}

void
debug_init( bool Daemonize )
{

    const char *log_file = NULL;

    if ( Daemonize ) {
        nest_debug_init( NULL, true );
    } else {
        nestProtocol *logfile = NULL;
        if( GetConfigValue( Config_LogFile, &log_file ) ) {
            assert( log_file != NULL );
            nest_printf( "Logging to %s.\n", log_file );
            logfile = storage_manager->openMetaFile( (char *)log_file, 
													 APPEND_FLAGS );
            assert( logfile != NULL );
        }
        if ( GetConfigValue( Config_DebugStderr, NULL ) ) {
            if ( logfile ) {
                nest_printf( "Logging to stderr and ignoring conflicting "
                        "directive to log to %s.\n", log_file );
            }
            logfile = NULL;
        }
        nest_debug_init( logfile, false );
    }
}

#define MAXFD 64

void
daemon_init( const char *pname, int facility )
{
    pid_t pid = 0;

    if ((  pid = fork() ) < 0 ) {
        nest_die( "fork failed: %s", strerror( errno ) );
    } else if ( pid > 0 ) {
            // parent
        exit( 0 );
    }

        // first child continues
    setsid();

    NestSignal( SIGHUP, SIG_IGN ); // ignore sig_pipes

        // forking twice is not working
    if ( ( pid = fork() ) <  0 ) {
        nest_die( "fork failed: %s", strerror( errno ) );
    } else if ( pid > 0 ) {
            // first child terminates
        exit( 0 );
    }

    umask( 0 );     // clear our file mode creation mask . . . ?

        // should we attempt to close all open descriptors . . . ?
    fprintf( stderr, "NeST has been daemonized.\n" );
    for( int i = 3; i < MAXFD; i++ ) {
        close( i );
    }
        // should we open /dev/null for reading and writing and dup it 
        // to stdin and stdout
    int devnull = NestOpen( "/dev/null", O_RDWR ); 
    NestDup2( devnull, STDIN_FILENO );
    NestDup2( devnull, STDOUT_FILENO );
        
    openlog( pname, LOG_PID | LOG_CONS, facility );
}

void 
taskDone( nestClient *client, NestReplyStatus status ) {

        // special case for transfers that put data but don't prespecify 
        // a transfer length.  we can only allocate lot space for these
        // transfers incrementally.
    if ( status == NEST_LOT_NOT_ENOUGH ) {
        status = storage_manager->increaseAllocation( client );
        if ( status == NEST_SUCCESS ) {
            status = concurrency_manager->addTask( client );
            return;
        } else {
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        }
    }


        // sometimes a proxy may add other special non-transfer requests
        // to the concurrency manager.  Allow them to clear up after them.
    if ( client->getTaskFunction() != NULL ) {
        taskCleanupFunction *cleanup = client->getTaskCleanupFunction();
        assert( cleanup != NULL );
        //NestReplyStatus cleanup_status = 
        cleanup( client, status );
        return;
    }

        // this persistent thing is for a weird case with http clients and
        // the pot concurrency.  It was possible for the taskDone to get
        // called after the client was removed.
        //
        // also however nfs clients are not persistent . . . 
        //
        // I think there still may be a race condition if the client object
        // is deleted while we're operating in taskDone
        // I think it is important to not call client->endTransfer until
        // ___after__ we've returned control to the server
        // because once client->endTransfer is called it is possible that
        // the client could be deleted.  So we can't reference it again.
        // 
    if ( client->hasSocket() && client->persistent() ){
        NestFd_set( client->getConnection()->getSocket(), &readfds );
        nest_debug( NEST_DEBUG_CONCURRENCY | NEST_DEBUG_FDSETS, 
                "Task_done adding %d to readfds %x\n", 
                client->getConnection()->getSocket(), &readfds );
    }



        // close the storage connections
    if ( client->getStorageConnection() != NULL ) {
        nestProtocol *storage_con = client->getStorageConnection();
        storage_manager->close( storage_con, client );
        delete( storage_con );
        client->assignStorageConnection( NULL ); 
    }

    if ( client->getOtherStorageConnection() != NULL ) {
        nestProtocol *storage_con = client->getOtherStorageConnection();
        storage_manager->close( storage_con ,client );
        delete( storage_con );
        client->assignOtherStorageConnection( NULL ); 
    }

        // TODO:  Be nice to maybe cache these.
    if ( client->getThirdPartyConnection() != NULL ) {
        nestProtocol *third = client->getThirdPartyConnection();
        delete( third );
        client->assignThirdPartyConnection( NULL );
    }
        // this seems weird to me (jb), why do we handle this one particular
        // error specially?
    if ( status == NEST_INSUFFICIENT_SPACE ) { 
        client->sendErrorMessage( __FILE__, __FUNCTION__, status);
    }
        // client endTransfer must be last thing to happen
        // otherwise, client might suddenly disappear
    client->endTransfer( status );

        // if we're running as proxy, start any who were waiting for this
        // and clean this one up
        // no longer needed, bec proxy taskDone is called directly 
        // above using the cleanup function pointer
    /*
    if ( serverIsProxy() ) {
        checkProxyTransfers( client, status ); 
    }
    */

}

int
HandleRequests( nestClientPool *requests, fd_set *readfds )
{
    nestClient *client;
    CLIENT_ITERATOR itr = requests->begin();
    while ( itr != requests->end() ) {
        client = *itr;
        itr = requests->erase( itr );

        switch( HandleRequest( client ) ) {
        case CLOSE_CONNECTION:
            RemoveClient( client, readfds );
            break;
        case PASSIVE_CONNECTION:
        case ACTIVE_CONNECTION:
                // nothing to do, we used to clear the client fd from the
                // read_fd set we pass to select here for an active
                // connection.  But now this is done in the concurrency
                // manager.
            break;
        default:
            assert( 0 );
            break;
        }
    } // end of client service loop
    return 0;
}

bool
RemoveClient( nestClient *client, fd_set *readfds )
{
    nest_debug( NEST_DEBUG_CLIENT, "Removing client %d\n", client->getId() );

        // if this client has a socket, stop listening to it 
    if ( client->hasSocket() ) {
        NestFd_clear( client->getConnection()->getSocket(), readfds);
    }

        // if this client registered as a proxy, remove them from proxy list
    if ( client->isProxy( ) ) {
        RemoveProxyClient( client );
    }

    if ( serverIsProxy() ) {
        removeClientFromProxy( client );
    }

    clients->removeClient( client );
    delete client;
    return true;
}

void
setNewGroup()
{
        // set ourselves up as a new group
    // doesn't work   arghh!
    //NestSetpgrp();
    /*
    if ( setpgid( 0, 37 ) ) {
        perror( "setpgid" );
    }
    if ( setsid() < 0 ) {
        perror( "setsid" );
    }
    */
}

int  
main_init( int argc, const char *argv[] )
{

    bool Daemonize = false;
    const char *value = NULL;

    set_nest_debug_level( NEST_DEBUG_MIN );
	nest_debug( NEST_DEBUG_ALWAYS,
				"NeST Server version '" NEST_VERSION __DATE__ "' starting\n" );

        // set up config file must occur before initing other stuff
    ReadConfigFile(argc, argv);

    setNewGroup();

        // set up our fd sets
    FD_ZERO( &writefds );
    FD_ZERO( &readfds  );

        // turn off all debugging unless instructed otherwise
    set_nest_debug_level( NEST_DEBUG_NEVER );

        // instantiate our client pool and our serverTasks
        // these need to be done before args, bec if we're a proxy,
        // we're gonna add the proxy monitor as a client
    clients     = new nestClientPool( storage_manager );
    serverTasks = new nestClientPool( storage_manager );


        // read config values
    if (GetConfigValue( Config_Debug, &value)) {
        if ( add_nest_debug_level(value) != NEST_SUCCESS ) {
            nest_die( "Unknown debug value specified: %s\n", value );
        }
    }

    if ( GetConfigValue( Config_PrintHelp, &value)) {
        PrintHelp(value);
        exit( 0 );
    }

    if ( GetConfigValue( Config_DumpConfigFile, NULL)) {
        DumpSampleConfigFile();
        exit( 0 );
    }

    if ( GetConfigValue( Config_Latency, &value ) ) {
        latency = atol( value );
        nest_printf( "Latency set at %ld ms.\n", latency );
    }

    if ( GetConfigValue(Config_MaintenanceInterval, &value ) ) {
        maintenance_interval = atoi(value);
    }

    if ( GetConfigValue( Config_ProcessUser, &value ) ) {
        if ( NestChangeUser( value ) < 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Unable to switch to user %s\n", value );
        }
    }

    if ( GetConfigValue( Config_Daemonize, NULL ) ) {
        Daemonize = true;
        daemon_init( "NeST", LOG_USER );
        //debug_init( true );
    }
    
    if ( GetConfigValue( Config_DisableAuthentication, NULL ) ) {
        allowedAuthentication = NEST_AUTH_NONE;
        nest_warn( __FILE__, __FUNCTION__, 
                "Running with authentication disabled is strongly "
                "discouraged!!\n" );
    }


    initCommonMutex();

        // set admin pid value, must be done after daemonizing
    admin_pid = NestGetpid();
    nest_debug( NEST_DEBUG_ADMIN, "Admin pid is %d\n", admin_pid );

        // debug_init must be called after storage and before concurrency
        //   unless Daemonize is true
        // protocolManager must proceed storage_manager
	storage_manager = new nestStorageManager( );
	identity_manager = new nestIdentityManager( storage_manager );
	file_catalog = new nestFileCatalog( storage_manager );
	lot_manager = new nestLotManager( storage_manager,
									  identity_manager,
									  file_catalog );
    collection_manager = new nestCollectionManager( storage_manager );

	// Ok, now initialize the storage manager, which will
	// initialize the others
	storage_manager->init( identity_manager,
						   file_catalog,
						   lot_manager,
						   collection_manager );

    /*if ( !Daemonize )*/ debug_init( false );

	// Create the protocol manager
	protocol_manager = new nestProtocolManager( storage_manager );
    protocol_manager->init( );

	// And, the advertiser
	advertiser = new nestAdvertiser( );

	// The concurrency manager
	concurrency_manager = new nestConcurrencyManager(
		 &readfds, 
		 &writefds, 
		 &r_readyfds, 
		 &w_readyfds,
		 taskDone,
		 storage_manager,
		 protocol_manager );
    concurrency_manager->init( );

	// Initialize the proxy logic
    initProxy( storage_manager, concurrency_manager,
			   TransferFile, clients, &readfds ); 

        // install signal handlers
    installSignalHandlers( );

        // add each available socket to our read set
        // the protocol manager has listened on one socket for each protocol
    for( int protocol_socket = protocol_manager->firstProtocolSocket();
         protocol_socket != protocol_manager->lastProtocolSocket();
         protocol_socket = protocol_manager->nextProtocolSocket() )
    {
        if ( protocol_socket > 0 ) {
			nest_debug( NEST_DEBUG_ALWAYS,
						"Adding socket %d (for %s) to read set\n",
						protocol_socket,
						protocol_manager->getProtocolNameFromSocket(protocol_socket ) );
            NestFd_set( protocol_socket, &readfds );
        }
    }

        // ask the ad manager to write the initial ad
    WriteAd();


    return 0;
}

// the max of the clients, the concurrencies and the protocols
// now that we add rpc fds to our select loop, I'm not sure how to set
// the maxfd . . . 
int
getMaxfd( )
{
    int maxfd = 
        max( concurrency_manager->getMaxfd( ), clients->getMaxfd());
    maxfd = max( maxfd, protocol_manager->maxSocket( ) );
    return maxfd;
}


NestReplyStatus
advertise( )
{
    NestReplyStatus status = NEST_SUCCESS;


        // update ad
    WriteAd();

# if 0
    char *host = "coral.cs.wisc.edu";
    int  port  = 9618;
    long len;
    nestProtocol *adfile = storage_manager->openMetaFile(AD_NAME,READ_FLAGS);
    len = storage_manager->metaFilelength( AD_NAME );
    nestProtocolAdvertise *collector = NULL;
    
    if ( collector == NULL ) {
        collector = new nestProtocolAdvertise( );
        if ( ! collector->connect( host, port ) ) {
            delete( collector );
            collector = NULL;
            nest_warn( __FILE__, __FUNCTION__,
					   "Unable to advertise to %s:%d\n", host, port );
            delete( adfile );
            return NEST_NO_CONNECTION;
        }
    }

        // create a fake client and assign it this transfer
    nestClient *client = new nestClient(
		storage_manager->getInitialClientDir(), collector, NULL );
    nestRequest *req   = new nestRequest(
		CLIENT_RECV_FILE, AD_NAME, NULL, NULL, 0, len, NULL );
    client->assignRequest( req );
    client->assignStorageConnection( adfile );

    serverTasks->addClient( client );
    status = concurrency_manager->addTask( client );
    nest_debug( NEST_DEBUG_SERVER,
            "Added advertise task to concurrencyMgr: %s\n", 
            NestErrorString( status ) );
    delete( adfile );
    return status;
# endif
            
    const char *condor_advertise;
    char command[NEST_MESSAGESIZE];
    int retval;

        // condor_advertise
    if ( GetConfigValue( Config_CondorAdvertise, &condor_advertise ) ) {
        snprintf( command, NEST_MESSAGESIZE, "%s UPDATE_STORAGE_AD %s",
            condor_advertise, storage_manager->metaFilepath( AD_NAME ) );
        retval = system( command );
        nest_debug( NEST_DEBUG_SERVER, "Exec'd system %s: %d\n",
                command, retval );
        if ( retval < 0 ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Unable to exec command \"%s\"\n", command );
        }
    }

    return status;
}


NestReplyStatus 
RegularMaintenance( )
{
    NestReplyStatus status = NEST_SUCCESS;

    nest_debug( NEST_DEBUG_SERVER, "Performing regular maintenance now\n" );
    NestGettimeofday( &last_maintenance );

        // only do concurrency::makeProgress for now
        // TODO need slow timeouts for regular maintenance
        // and fast timeouts for scheduling
        // turn off anticipatory scheduling for now

        // for a while with the proxies, this was needed bec the pot pipe
        // was somehow maybe not waking up the server
        // but I think that was a race condition otherwise solved.
    //concurrency_manager->makeProgress();

    advertise();

    storage_manager->maintenance();

    proxyMaintenance();

    return status;
}

/*
void
sigInt( int signal ) {
    nest_printf( "Exiting due to SIGINT.\n" );
    _exit( 0 );
}
*/

// ignore sigpipes
// pass sigchilds to the concurrency manager
void
installSignalHandlers( )
{

    NestSignal( SIGPIPE, SIG_IGN ); // ignore sig_pipes
 //   NestSignal( SIGINT,  sigInt );

    struct sigaction sa;
    sa.sa_sigaction = sigChild;
    sigemptyset( &sa.sa_mask );
    sa.sa_flags = SA_SIGINFO;

    if ( sigaction( SIGCHLD, &sa, 0 ) != 0 ) {
        nest_die( "sigaction: %s", strerror( errno ) );
    }
    if ( sigaction( SIGCLD, &sa, 0 ) != 0 ) {
        nest_die( "sigaction: %s", strerror( errno ) );
    }
}

void
sigChild( int signal, siginfo_t *info, void *context )
{
    concurrency_manager->sigChild( signal, info, context );    
}

nestRequest *
ReceiveRequest( nestClient *client, fd_set *ready )
{
    struct timeval start, end;
    nestRequest *req;

    NestGettimeofday( &start );
    req = client->receiveRequest( ready );
    NestGettimeofday( &end );
    if ( NestTimediff( &start, &end, __FUNCTION__ ) > 1000000 ) {
        nest_warn(
			__FILE__, __FUNCTION__,
			"Receive request took %.3f secs.\n", 
			(float)NestTimediff( &start, &end, __FUNCTION__ ) / 1000000 );
    }
	if ( req && (! req->isValid() ) ) {
		nest_warn( __FILE__, __FUNCTION__,
				   "Received invalidated request %d (%s) from client %s\n",
				   req->getType(), req->toString(), client->toString() );
	}
    return req;
}

int
HandleRequest( nestClient *client )
{
	nestRequest	*request = client->getRequest( );
    switch( request->getType() ) {
    case CLIENT_RECV_FILE:
    case CLIENT_SEND_FILE:
    case THIRD_PARTY_GET:
    case THIRD_PARTY_PUT:
    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
        return TransferFile( client, request );
    case COPY_REQUEST:
        return CopyFile(client, request );
    case CLOSE_CONNECTION_REQUEST:
        client->endRequest( NEST_SUCCESS );
        return CLOSE_CONNECTION;
    case NEST_SYNC:
        return NestSync( client, request );
    default:
        return SynchronousRequestHandler( client, request );
    }

}

// these quick simple requests are just handled immediately by the main server
int
SynchronousRequestHandler( nestClient *client, nestRequest *request )
{

    int connection_state = PASSIVE_CONNECTION;
    nest_debug( NEST_DEBUG_SERVER, "SynchronousRequestHandler for req %d\n",
            request->getType() );
    NestReplyStatus status = NEST_SUCCESS;

        // respond appropriately to each request type
    switch( request->getType() ) {
    case PROXY_REGISTER:
        status = RegisterProxyClient( client, request );
        break;
    case PROXY_QUERY:
        status = QueryProxys( client, request );
        break;
    case PROXY_FLUSH:
        status = FlushProxyCache( client );
        break;
    case PROXY_INDEX_PUSH:
        status = ReceiveProxyIndex( client );
        break;
    case FORMAT_DISK_REQUEST:
        status = FormatDisk( client, request );
        break;
    case BAD_REQUEST:
        status = client->sendErrorMessage( __FILE__, __FUNCTION__,  
                request->getStatus() );
        break;
    case VERSION_REQUEST:
        status = SendVersion( client, request );
        break;
    case REQUEST_NEST_AD:
        status = SendNestAd( client, request );
        break;
    case GET_FILEHANDLE:
        status = GetFilehandle( client, request );
        break;
    case AVAILABLE_SPACE_QUERY:
        status = SendAvailableSpace( client, request );
        break;
    case TIME_REQUEST:
        status = client->sendTimestamp( );
        break;
    case CONNECTION_REQUEST:
    case CHANGE_USER_REQUEST:
        status = AuthenticateUser( client, request );
        break;
    case ADD_NEST_USER:
        status = AddUser( client, request );
        break;
    case ADD_NEST_GROUP:
        status = AddGroup( client, request );
        break;
    case DELETE_NEST_USER:
        status = DeleteUser( client, request );
        break;
    case DELETE_NEST_GROUP:
        status = DeleteGroup( client, request );
        break;
    case ADD_USER_TO_GROUP:
        status = AddUserToGroup( client, request );
        break;
    case DELETE_USER_FROM_GROUP:
        status = DeleteUserFromGroup( client, request );
        break;
    case REQUEST_USER_LOT:
        status = ReserveUserSpace( client, request );
        break;
    case REQUEST_GROUP_LOT:
        status = ReserveGroupSpace( client, request );
        break;
    case LIST_USER_LOTS:
        status = ListUserLots( client, request );
        break;
    case LIST_GROUP_LOTS:
        status = ListGroupLots( client, request );
        break;
    case GET_LOT_STAT:
        status = GetLotStat( client, request );
        break;
    case UPDATE_LOT:
        status = UpdateLot( client, request );
        break;
    case CHANGE_LOT_T:
        status = ChangeLotType( client, request );
        break;
    case LOT_MOVE_FILE:
        status = MoveFileLot( client, request );
        break;
    case ADD_USER_TO_LOT:
        status = AddUserToLot( client, request );
        break;
    case DELETE_USER_FROM_LOT:
        status = DeleteUserFromLot( client, request );
        break;
    case ATTACH_LOT_TO_FILE:
        status = AttachLotToFile( client, request );
        break;
    case DETACH_LOT_FROM_FILE:
        status = DetachLotFromFile( client, request );
        break;
    case TERMINATE_LOT:
        status = TerminateLot( client, request );
        break;
    case FILE_DELETE_REQUEST:
        status = DeleteFile( client, request );
        break;
    case FILE_RENAME_REQUEST:
        status = RenameFile( client, request );
        break;
    case BANDWIDTH_STATISTICS_REQUEST:
        status = SendBandwidthStatistics( client, request );
        break;
    case DIR_LIST_REQUEST:
        status = SendDirectoryList( client, request );
        break;
    case FILEINFO_REQUEST:
        status = SendFileinfo( client, request );
        break;
    case SETATTR_REQUEST:
        status = SetFileAttr( client, request );
        break;
    case FILESIZE_REQUEST:
        status = SendFilesize( client, request );
        break;
    /** This has been replaced by the copy function that uses concurrency
        And so it appears along with the other transfer functions
    case COPY_REQUEST:
        status = CopyFile( client, request );
        break;
    */
    case LINK_REQUEST:
        status = MakeLink( client, request );
        break;
    case PWD_REQUEST:
        status = SendPwd( client, request );
        break;
    case MKDIR_REQUEST:
    case CWD_REQUEST:
    case RMDIR_REQUEST:
        status = DirOp( client, request );
        break;
    case PHYSNAME_REQUEST:
        nest_debug( NEST_DEBUG_SERVER, "PhysName request\n" );
        status = SendPhysName( client, request );
        break;
    case GET_ACL:
    case GET_GROUP_ACL:
        nest_debug( NEST_DEBUG_SERVER, "GetUserAcl request\n" );
        status = SendUserAcl( client, request );
        break;
    case SET_ACL:
    case SET_GROUP_ACL:
        nest_debug( NEST_DEBUG_SERVER, "SetUserAcl request\n" );
        status = SetUserAcl( client, request );
        break;
    case NAME_COLLECTION:
    case UNNAME_COLLECTION:
    case QUERY_COLLECTION:
	case SET_COLLECTION_MTIME:
	case SET_COLLECTION_VERSION:
	case SET_COLLECTION_COMMENT:
	case GET_COLLECTION_MTIME:
	case GET_COLLECTION_VERSION:
	case GET_COLLECTION_COMMENT:
		status = CollectionOp( client, request );
        WriteAd();
        RegularMaintenance();
        break;
    case SERVER_TERMINATE_REQUEST:
        if ( TerminateServer( client, request ) == NEST_SUCCESS ) {
            exit( 0 );
        }
        break;
    case DEBUG_REQUEST:
        status = DebugRequest( client, request );
        break;
    default:
        assert( 0 );
        return CLOSE_CONNECTION;
    }

    client->endRequest( status );

    if ( connection_state == PASSIVE_CONNECTION ) {
            // this means the server is responsible for listening to the client
        if ( ! client->persistent() ) {
                // client is a one time visitor (e.g. nfs or http)
            connection_state = CLOSE_CONNECTION;
        }
    }
    return connection_state;
}

NestReplyStatus
TerminateServer( nestClient *client, nestRequest *request )
{

    NestReplyStatus status = NEST_NONPRIVELEDGED_USER;
    const char *tmp_admin, *admin;

        // check whether user is the registered admin user
    if ( ! GetConfigValue( Config_AdminUser, &tmp_admin ) ) {
        status = NEST_NONPRIVELEDGED_USER;
    } else {
        admin = NestGetModifiedIdentity(tmp_admin);
        if ( !strcmp( client->getUserIdentity(), admin ) ) {
            status = NEST_SUCCESS;
        } else {
            status = NEST_NONPRIVELEDGED_USER;
        }
    }

    if ( status == NEST_SUCCESS ) {
        client->sendAck();
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }

    return status;
}

NestReplyStatus
DebugRequest( nestClient *client, nestRequest *req )
{
	const char	*string = req->getDebugString( );

    NestReplyStatus status = add_nest_debug_level( string );
    if ( status == NEST_SUCCESS ) {
        concurrency_manager->debugLevel( get_nest_debug_level() );
        nest_debug( NEST_DEBUG_DEBUG, "nest_debug set to %s\n", string );
        client->sendAck();
    } else {
        nest_debug( NEST_DEBUG_DEBUG, "could not set nest_debug set %s\n",
            string );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
WriteAd( )
{
    NestReplyStatus status;

    nestProtocol *ad = storage_manager->openMetaFile
        ( AD_NAME, WRITE_FLAGS | O_TRUNC );

    if ( ad == NULL ) {
        nest_die( "Unable to open metafile %s\n", AD_NAME );
        status = NEST_NONPRIVELEDGED_SERVER;
    } else {
		// status = ad->Sendn( "[\n", 2 ); this is new classads
        status = advertiser->writeAd( storage_manager,
									  collection_manager,
									  concurrency_manager,
									  protocol_manager,
									  ad );

            // now write some server stuff
        char buffer[NEST_MESSAGESIZE];
        snprintf( buffer, NEST_MESSAGESIZE, 
                "Current_clients   = %d\n"
                "Maintenance_interval = %.2f\n",
                ( ( clients != NULL ) ? clients->size() : 0 ),
                (float)maintenance_interval / THOUSAND );

        if ( serverIsProxy() ) {
            writeProxyAd( ad );
        }

		// this is new classads
		//snprintf( &buffer[NestStrlen(buffer)], NEST_MESSAGESIZE, "]\n" ); 
        snprintf( &buffer[NestStrlen(buffer)], NEST_MESSAGESIZE, "\n" ); 
        
        ad->Sendn( buffer, NestStrlen( buffer ) );
        delete( ad );
    }

    return status;
}

NestReplyStatus 
GetFilehandle( nestClient *client, nestRequest *request )
{
    NestFileInfo info;

    NestReplyStatus status = storage_manager->fileinfo( 
            &info, 
            client->getRequest()->getNestFile(),
            client->getUser( ),
            client->getCwd() );

    if ( status == NEST_SUCCESS ) {
        client->sendFilehandle( &info );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
SendFileinfo( nestClient *client, nestRequest *request )
{

    NestFileInfo	info;

	info.next = NULL;
	info.classad = NULL;

    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    if ( serverIsProxy() ) {
        status = proxyFileinfo(
            &info,
            client->getRequest()->getNestFile(),
            client->getUser(),
            client->getCwd() );
    } else {
        status = storage_manager->fileinfo(
            &info,
            client->getRequest()->getNestFile(),
            client->getUser(),
            client->getCwd() );

    }

    if ( status == NEST_SUCCESS ) {
        client->sendFileinfo( &info );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
SetFileAttr( nestClient *client, nestRequest *request )
{
    NestReplyStatus     retval;
    char                *phys_name;

    retval = storage_manager->get_physical_filename(
        &phys_name,
        request->getFile(),
        client->getUser(),
        client->getCwd() );
    if( retval != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }
   
    NestFileInfo info; 
    retval = storage_manager->fileinfo( 
            &info, 
            request->getFile(),
            client->getUser(),
            client->getCwd() );
    if( retval != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }
      
    client->setFileattr(&info, 
                        phys_name,
                        request->getFileAttr() );
    NestFree( phys_name );

    return retval;
}

NestReplyStatus
SendFilesize( nestClient *client, nestRequest *request )
{

    nestFileSize filesize;

    NestReplyStatus status = NEST_NOT_YET_INITIALIZED;
    if ( serverIsProxy() ) {
        status = proxyFilelength(
            &filesize,
            client->getRequest()->getNestFile(),
            client->getUser(),
            client->getCwd() );
    } else {
        status = storage_manager->filelength(
            &filesize,
            client->getRequest()->getNestFile(),
            client->getUser(),
            client->getCwd() );
    }

    if ( status == NEST_SUCCESS ) {
        client->sendFilesize( filesize );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

NestReplyStatus
SendPhysName( nestClient *client, nestRequest *request )
{

    NestReplyStatus        retval;
    char                *phys_name;

    retval = storage_manager->get_physical_filename(
        &phys_name,
        client->getRequest()->getNestFile(),    
        client->getUser(),
        client->getCwd() );

    if ( retval == NEST_SUCCESS ) {
        client->sendPhysName( phys_name );
        NestFree( phys_name );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
    }

    return retval;
}

NestReplyStatus
SendUserAcl( nestClient *client, nestRequest *request )
{
    NestReplyStatus				retval;

	// Get the user/group
	const nestUserOrGroup		*target_ug;
    if ( request->getType() == GET_ACL ) {
		target_ug = request->getUser( );
	} else {
		target_ug = request->getGroup( );
	}

    // Target user specified?
    nest_debug( NEST_DEBUG_SERVER,
                "SendUserAcl: User/Group=%p='%s' Group=%c, File='%s'\n",
				target_ug->getName(),
				( request->getType() == GET_ACL ) ? 'f' : 't',
                client->getRequest()->getNestFile() );

    // Check for all users?
	if ( NULL == target_ug ) {
        NestACL    *Acl;

        retval = storage_manager->get_acl(
                    &Acl,
                    client->getRequest()->getNestFile(),
                    client->getUser(),
                    client->getCwd() );
        nest_debug( NEST_DEBUG_SERVER,
                    "SendUserAcl[ACL]: retval=%d, Acl @@ %p\n", retval, Acl );
        if ( retval == NEST_SUCCESS ) {
            client->sendAcl( Acl );
        } else {
            client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
        }
    } else {
        // Just a single (specified) user -or- group
        nest_access_rights_t    access;
		retval = storage_manager->get_perm(
			&access,
			target_ug,
			client->getRequest()->getNestFile(),
			client->getUser(),
			client->getCwd() );
        nest_debug( NEST_DEBUG_SERVER,
                    "SendUserAcl[User='%s']: Access=0x%02x\n", 
                    target_ug->getName(), access );
        if ( retval == NEST_SUCCESS ) {
            client->sendAcl( target_ug->getName(),
							 target_ug->isGroup(),
							 access );
        } else {
            client->sendErrorMessage( __FILE__, __FUNCTION__,  retval );
        }
    }

    return retval;
}

NestReplyStatus
SetUserAcl( nestClient *client, nestRequest *request )
{
    NestReplyStatus				status;

	// Get the user/group
	const nestUserOrGroup		*target_ug;
    if ( request->getType() == GET_ACL ) {
		target_ug = request->getUser( );
	} else {
		target_ug = request->getGroup( );
	}

    // Get request parameters
    nest_debug( NEST_DEBUG_SERVER,
				"SetUserAcl: User='%s', IsGroup=%c, File='%s', ACL=0x%02x\n",
				target_ug->getName(),
				target_ug->isGroup() ? 't' : 'f',
                request->getFile(),
				request->getRights() );

    // Make the check
	status = storage_manager->set_perm( request->getRights(),
										target_ug,
										request->getFile(),
										client->getUser(),
										client->getCwd() );

    nest_debug( NEST_DEBUG_SERVER, "SetUserAcl: status=%d\n", status );

    // Done
    if ( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    return status;
}

NestReplyStatus
SendVersion( nestClient *client, nestRequest *request )
{
    client->sendVersion( NEST_VERSION );
    return NEST_SUCCESS;
}


NestReplyStatus
SendPwd( nestClient *client, nestRequest *request )
{
    string pwd;

    NestReplyStatus reply = storage_manager->expand_path( 
            pwd, client->getUser(),
            client->getCwd(), "" );

    if (reply != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__, reply);
    } else {
        client->sendPwd( pwd.c_str() );
    }

    return reply;
}

NestReplyStatus
CollectionOp( nestClient *client, nestRequest *request )
{
	string name = request->getCollectionName( );
	bool			sendack = false;

    NestReplyStatus status = NEST_SUCCESS;
    nest_obj_type_t filetype;

    switch( request->getType() ) {

    case NAME_COLLECTION:
        status = storage_manager->get_type( &filetype,
											request->getFile( ),
											client->getUser( ),
											client->getCwd() );

        if (status == NEST_SUCCESS ) {
            string pathname;
            storage_manager->expand_path( pathname,
										  client->getUser( ),
										  client->getCwd(),
										  request->getFile( ) );

            status = collection_manager->createAssociation( name, pathname );
        }
		sendack = true;
        break;

    case UNNAME_COLLECTION:
        status = collection_manager->removeAssociation( name );
		sendack = true;
        break;

	case SET_COLLECTION_MTIME:
	case SET_COLLECTION_VERSION:
	case SET_COLLECTION_COMMENT:
	case GET_COLLECTION_MTIME:
	case GET_COLLECTION_VERSION:
	case GET_COLLECTION_COMMENT:
	{
		sendack = true;
		nestCollection	*col = collection_manager->getCollection( name );
		if ( ! col ) {
			return NEST_LOCAL_FILE_NOT_FOUND;
		}

		switch( request->getType() ) {
		case SET_COLLECTION_MTIME: {
			col->setMtime( request->getCollectionMtime() );
			break;
		}
		case SET_COLLECTION_VERSION: {
			col->setVersion( request->getCollectionVersion( ) );
			break;
		}
		case SET_COLLECTION_COMMENT: {
			col->setComment( request->getCollectionComment( ) );
			break;
		}
		case GET_COLLECTION_MTIME:
		case GET_COLLECTION_VERSION:
		case GET_COLLECTION_COMMENT:
			break;
		}
		delete col;
		break;
	}

    case QUERY_COLLECTION:
    default:
        assert( 0 );
    }

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else if ( sendack ) {
		client->sendAck( );
	} else {
		assert( 0 );		// QUERY_COLLECTION not yet in chirp
    }

    return status;
}

NestReplyStatus
DirOp( nestClient *client, nestRequest *request )
{
    const char *dir  = request->getDir();
    const nestUser *user = client->getUser();
    const char *cwd  = client->getCwd();

    NestReplyStatus status;

    switch( request->getType() ) {
    case RMDIR_REQUEST:
        status = storage_manager->rmDir( user, cwd, dir );
        break;

    case CWD_REQUEST:
	{
		char *new_cwd;
        status = storage_manager->chDir( &new_cwd, user, cwd, dir );
        if (status == NEST_SUCCESS) {
            status = client->setCwd( new_cwd );
            NestFree( new_cwd );
        }
        break;
	}

    case MKDIR_REQUEST:
        status = storage_manager->mkDir( user, cwd, dir, user->getUid() );
        break;

    default:
        assert( 0 );
    }

    if ( status == NEST_SUCCESS ) {
        client->sendAck( );
    }
    else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

void
getAuthString( int types, string &str )
{
	static	struct { int	mask; const char *name; } table [] = {
		{ NEST_AUTH_NONE, "none" },
		{ NEST_AUTH_GSS, "GSS" },
		{ NEST_AUTH_IP, "IP" },
		{ 0, NULL }
	};

	// Walk through the table..
	int		count = 0;
	for ( int i = 0;  table[i].mask;  i++ ) {
		if ( types & table[i].mask ) {
			if ( count++ ) {
				str += ", ";
			}
			str += "\"";
			str += table[i].name;
			str += "\"";
		}
	}
}

NestReplyStatus
AuthenticateUser( nestClient *client, nestRequest *request )
{
    char *username = NULL;
    char *identity = NULL;
    NestReplyStatus stat = NEST_SUCCESS;


	int		auth_type = request->getAuthType();
	string	auth_str;
	getAuthString( auth_type, auth_str );

	// Note: This is not-reentrant
	static string	allowed_str;
	static int		allowed_cache = -1;
	if ( allowed_cache != allowedAuthentication ) {
		allowed_str = "";
		allowed_cache = allowedAuthentication;
		getAuthString( allowedAuthentication, allowed_str );
	}

        // check whether requested authentication method is allowed
    if ( ! ( auth_type & allowedAuthentication) ) {
        nest_debug( NEST_DEBUG_AUTH,
					"Client %s requested unallowed authentication type:"
					" %d %s\n",
					request->getUser()->getName(),
					auth_type, auth_str.c_str() );
		nest_debug( NEST_DEBUG_AUTH, "Allowed types: %s\n",
					allowed_str.c_str() );
        stat = NEST_INVALID_AUTHENTICATION;
        client->sendErrorMessage( __FILE__, __FUNCTION__,  stat ); 
        return stat;

    } else if ( check_nest_debug_status( NEST_DEBUG_AUTH ) ) {
		const nestUser *user = request->getUser();
        nest_debug( NEST_DEBUG_AUTH,
					"Client %s (%s) requesting auth type %d %s\n",
					user->getName(), client->getHostIp(),
					auth_type, auth_str.c_str() );
    }

        // check requested authentication method
    switch( auth_type ) {

    case NEST_AUTH_GSS:
        stat = client->getConnection()->gssAccept( &username, &identity );
        if ( stat != NEST_SUCCESS ) {
            nest_debug( NEST_DEBUG_AUTH,
						"Client %s (%s) couldn't authenticate: %s\n",
						request->getUser(),
						client->getHostIp(),
						NestErrorString( stat ) );
            client->sendErrorMessage( __FILE__, __FUNCTION__,  stat );
        }
        break;

    case NEST_AUTH_NONE:
        username = NestStrdup( (char*)request->getUser() );
        identity = NestStrdup( username );
        break;

	case NEST_AUTH_IP:
		// TODO:  put in IP based authentication ala Kangaroo
        stat = NEST_INVALID_AUTHENTICATION;
		client->sendErrorMessage( __FILE__, __FUNCTION__,  stat );
		nest_warn( __FILE__, __FUNCTION__,
				   "Client %s (%s) requested IP authentication"
				   " which is not implmented\n",
				   request->getUser(), client->getHostIp() );
		break;

    default:
        nest_warn( __FILE__, __FUNCTION__, "Unknown auth type requested\n" );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  
                NEST_INVALID_AUTHENTICATION );
        stat = NEST_INVALID_AUTHENTICATION;
    }

        // if we make it here, the authentication was successful
        // now go ahead and set the user name accordingly
    if ( stat == NEST_SUCCESS ) {
        assert( username != NULL );
        assert( identity != NULL || (allowedAuthentication & NEST_AUTH_NONE) );

		// Create a nest user object
		nestUser	*user = new nestUser( identity );
		user->setLocalName( username );

		int valid = storage_manager->validate_user( user );
		if ( valid ) {

            nest_debug( NEST_DEBUG_SERVER | NEST_DEBUG_CLIENT,
						"Client identity set to %s (user %s / uid %d)\n", 
						user->getIdentity(),
						user->getOsName(),
						user->getUid() );

			// Create user home directory...
			(void) storage_manager->add_home_dir( user );

			// Note: client->setUser() passes ownership of the user object
			client->setUser( user );
			user = NULL;
            client->sendAck( );
        } else {
            nest_warn( __FILE__, __FUNCTION__, 
                    "WARN: unable to change to user: %s\n", identity );
            client->sendErrorMessage(__FILE__, __FUNCTION__,
									 NEST_UNKNOWN_USER);
        }

        if ( username != NULL ) {
			NestFree( username );
		}
        if ( identity != NULL ) {
			NestFree( identity );
		}
        if ( user != NULL ) {
			delete user;
		}

            // proxy servers may need to do something special with
            // this client
        if ( serverIsProxy() ) {
            addNewClientToProxy( client ); 
        }

    }

    return stat;
}

NestReplyStatus
AddUser( nestClient *client, nestRequest *request )
{
    NestReplyStatus status;

	// Create a user object with the default group
	const nestGroup	*new_group = request->getGroup( );
	nestUser		new_user( request->getUser( ) );
	new_user.setGroup( new_group );

	status = storage_manager->add_user( client->getUser(), &new_user );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
DeleteUser( nestClient *client, nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->delete_user( client->getUser(),
										   request->getUser() );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
AddGroup( nestClient *client, nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->add_group( client->getUser(),
										 request->getGroup() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
DeleteGroup( nestClient *client, nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->delete_group( client->getUser(),
											request->getGroup() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
AddUserToGroup( nestClient *client, nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->add_user_to_group(
		client->getUser(),
		request->getUser(),
		request->getGroup() );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
DeleteUserFromGroup( nestClient *client, nestRequest *request )
{
    NestReplyStatus status;

    status = storage_manager->delete_user_from_group(
		client->getUser(), request->getUser(), request->getGroup() );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
ReserveUserSpace( nestClient *client, nestRequest *request )
{
    NestReplyStatus status = NEST_SUCCESS;
    NestLotHandle handle;

	// Create a "nest user" object
	const nestUser *user = request->getUser( );
	if( ! user ) {
		user = client->getUser( );
	}

	// And, one for the parent user / group
	status = lot_manager->request_user_lot(
		&handle,
		client->getUser( ),
		request->getLotSize( ),
		request->getLotDuration( ),
		user,
		request->getLotParentUG( ),
		request->getLotParentLot( ),
		request->getLotType( )
		);

	{
		rsv_id_t	tmpid;
		NestGet_id_from_handle( &handle, &tmpid );
		nest_debug( NEST_DEBUG_SERVER, "ReserveUserSpace: Handle=%p, id=%d\n",
					&handle, tmpid );
	}
                        
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
        client->sendLotHandle( &handle );
    }
    return status;
}

NestReplyStatus
ReserveGroupSpace( nestClient *client, nestRequest *request )
{
    NestReplyStatus status;
    NestLotHandle handle;

    status = lot_manager->request_group_lot(
                &handle,
                client->getUser(),
                request->getLotSize(),
                request->getLotDuration(),
                request->getGroup( ),
				request->getLotParentUG( ),
                request->getLotParentLot( ),            
                request->getLotType( )
             );
            
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
        client->sendLotHandle( &handle );
    }
    return status;
}

NestReplyStatus
ListUserLots( nestClient *client, nestRequest *request )
{
    NestLotListInternal     *result  = NULL;
    NestReplyStatus status;

	const nestUser	*user = request->getUser( );
	if ( ! user ) {
		user = client->getUser( );
	}

	status = lot_manager->list_lots(
		&result,
		client->getUser(),
		user );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        while (result) {
            NestLotListInternal *tmp = result->next;
            delete result;
            result = tmp;
        }
        return status;
    } else {
        client->sendAck();
        while (result) {
            NestLotListInternal *tmp = result->next;
            client->sendNestLot( &result->stat );
            delete result;
            result = tmp;
        }
    }

    client->sendAck();
    return status;
}

NestReplyStatus
ListGroupLots( nestClient *client, nestRequest *request )
{
    NestLotListInternal     *result = NULL;
    NestReplyStatus status;

	status = lot_manager->list_lots(
		&result,
		client->getUser( ),
		request->getGroup( ) );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        while (result) {
            NestLotListInternal *tmp = result->next;
            delete result;
            result = tmp;
        }
        return status;
    } else {
        client->sendAck();
        while (result) {
            NestLotListInternal *tmp = result->next;
            client->sendNestLot( &result->stat );
            delete result;
            result = tmp;
        }
    }

    client->sendAck();
    return status;
}

NestReplyStatus
GetLotStat( nestClient *client, nestRequest *request )
{
    NestReplyStatus ret_val;
    NestLotInternal rstat;

    const NestLotHandle   *handle = request->getLotHandle( );
    if ( ! handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    ret_val = lot_manager->stat_lot(
		handle,
		client->getUser( ),
		&rstat );

    if (ret_val != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  ret_val );
    } else {
        client->sendAck();
        client->sendNestLot( &rstat );
    }
    return ret_val;
}

NestReplyStatus
UpdateLot( nestClient *client, nestRequest *request )
{
	NestReplyStatus status;

	const NestLotHandle   *handle = request->getLotHandle( );
    if ( ! handle ) {
        status = NEST_BAD_PARAMETERS;
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

    status = lot_manager->update_lot(
		handle,
		client->getUser( ),
		request->getLotSize( ),
		request->getLotDuration() );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
ChangeLotType( nestClient *client, nestRequest *request )
{
    NestReplyStatus status;

	const NestLotHandle   *handle = request->getLotHandle( );
     status = lot_manager->change_lot_type(
                handle,
                client->getUser( ),
                request->getLotType(),
                request->getLotDuration() );
    
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
MoveFileLot( nestClient *client, nestRequest *request )
{
    NestReplyStatus status;
    nestLotSize movebytes = request->getLotBytesToMove();
    const NestLotHandle *source_lot = request->getLotSourceHandle( );
	const NestLotHandle *dest_lot = request->getLotDestHandle( );

    // Extract the file name..

    // Check the Bytes to move
    if (movebytes < 0 || !source_lot || !dest_lot )
    {
        status = NEST_BAD_PARAMETERS; 
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        return status;
    }

    status = storage_manager->modify_lot_usage(
        client->getUser( ),
        request->getNestFile(),
        client->getCwd(),
        source_lot,
        movebytes,
        dest_lot );

    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
TerminateLot( nestClient *client, nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );

    NestReplyStatus status;

    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    status = lot_manager->terminate_lot( handle,
										 client->getUser() );
    if ( status != NEST_SUCCESS ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
AddUserToLot( nestClient *client, nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );
    NestReplyStatus status; 
        
    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    status = lot_manager->add_user_to_lot( handle,
										   client->getUser( ),
										   request->getUser( ) );

    if(status != NEST_SUCCESS) {    
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;
}

NestReplyStatus
DeleteUserFromLot( nestClient *client, nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );
    NestReplyStatus status; 
        
    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    status = lot_manager->delete_user_from_lot(
		handle,
		client->getUser( ),
		request->getUser() );

    if(status != NEST_SUCCESS) {    
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }
    return status;    
}

NestReplyStatus
AttachLotToFile( nestClient *client, nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );

    NestReplyStatus status;

    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,
								  NEST_BAD_PARAMETERS );
    }

    status = storage_manager->attach_lot_to_file( handle,
												  request->getFile( ),
												  client->getUser( ),
												  client->getCwd() );


    if(status != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }

    return status;
}                                 
                                                    
NestReplyStatus
DetachLotFromFile( nestClient *client, nestRequest *request )
{
    const NestLotHandle *handle = request->getLotHandle( );

    NestReplyStatus status;

    if ( !handle ) {
        client->sendErrorMessage( __FILE__, __FUNCTION__, 
								  NEST_BAD_PARAMETERS );
    }

    status = storage_manager->detach_lot_from_file( handle,
													request->getFile( ),
													client->getUser( ),
													client->getCwd() );

    if(status != NEST_SUCCESS) {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        client->sendAck();
    }

    return status;
}            

NestReplyStatus
MakeLink( nestClient *client, nestRequest *request )
{
    NestReplyStatus status = storage_manager->Link(
            client->getUser( ),
            request->getFile(),
            request->getDestFile(),
            request->getIsSymLink(),
            client->getCwd() );

    if ( status == NEST_SUCCESS ) {
        client->sendAck();
    }
    else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }

    return status;
}

NestReplyStatus
CopyFile( nestClient *client, nestRequest *request )
{
	// This copy function in storage manager is no longer used

#if 0
    NestReplyStatus status =
        storage_manager->copyFile( client->getIdentity(),
                                      request->getNestFile(),
                                      request->getLinkName(),
                                      client->getCwd() );

    if( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    return status;
#endif

    NestReplyStatus status;

    nestProtocol *storage, *otherStorage;

    status = openStorageConnections( &storage, &otherStorage, client );
    if(status == NEST_SUCCESS){
        client->assignStorageConnection( storage );
        client->assignOtherStorageConnection( otherStorage );
        status = concurrency_manager->addTask( client );
    }
        
    if ( status == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_TRANSFER,
                "copy transfer added to concurrency manager\n" );
        client->dump( );
        return ACTIVE_CONNECTION;
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        client->endRequest( status );
        nest_debug( NEST_DEBUG_TRANSFER, "copy transfer failure: %s\n",
                NestErrorString( status ) );
        client->dump( );
        return PASSIVE_CONNECTION;
    }

}

NestReplyStatus
DeleteFile( nestClient *client, nestRequest *request )
{

    const char *file = request->getNestFile();

    NestReplyStatus status = 
        storage_manager->deleteFile( client->getUser( ), 
									 file,
									 client->getCwd() );

    if ( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );

    return status;
}

NestReplyStatus
RenameFile( nestClient *client, nestRequest *request )
{
    NestReplyStatus status =
        storage_manager->renameFile( client->getUser( ),
									 request->getFile(),
									 request->getDestFile(),
									 client->getCwd()
									 );
                                
    if ( status == NEST_SUCCESS ) client->sendAck( );
    else client->sendErrorMessage( __FILE__, __FUNCTION__,  status );

    return status;
    
}


NestReplyStatus
openStorageConnections( nestProtocol **storage,
                        nestProtocol **otherStorage,
                        nestClient   *client)
{

    NestReplyStatus     reply;
    nestProtocol *storage_con, *other_storage_con;
    nestRequest *request = client->getProxyRequest();
    nestFileSize offset = request->getStart();
	int importance = 1;
	int link_count = 0;
    
    nestFileSize filesize;
    reply = storage_manager->filelength(
		&filesize,
		request->getNestFile(),
		client->getUser( ),
		client->getCwd() );
    NEST_REPLY_CHECK(reply);
    request->setEnd(filesize);

	// Open the destination file
    reply = storage_manager->open( &storage_con,
								   client->getUser(),
								   request->getDestFile(),
								   WRITE_FLAGS,
								   offset,
								   filesize,
								   client->getCwd(),
								   link_count,
								   importance );
	*storage = ( reply == NEST_SUCCESS ) ? (nestProtocol *)storage_con : NULL;
    NEST_REPLY_CHECK(reply);

	// Open the source file
    reply = storage_manager->open( &other_storage_con,
								   client->getUser(),
								   request->getFile(),
								   READ_FLAGS,
								   offset,
								   filesize,
								   client->getCwd(),
								   link_count,
								   importance );
	*otherStorage =
		( reply == NEST_SUCCESS ) ? (nestProtocol *)other_storage_con : NULL;
    NEST_REPLY_CHECK(reply);

    nest_debug(NEST_DEBUG_TRANSFER, "Created copy transfer: %s->%s\n",
			   request->getNestFile(), request->getDestFile() );

    return reply;
}

NestReplyStatus
openStorageConnection( nestProtocol **storage,
					   nestClient *client,
					   const nestRequest *request )
{

    NestReplyStatus     reply = NEST_SUCCESS;
    nestFileSize  offset;
    nestFileSize filesize;

	int importance = 1;
	int link_count = 0;
	nestProtocol *tmp;

        // TODO: check with consistency manager
    switch( request->getType() ) {
    case CLIENT_SEND_FILE:
    case THIRD_PARTY_GET:
		if ( serverIsProxy() ) {
				// don't delete explicitly put files
			importance = INFINITE_IMPORTANCE;
			nest_debug( NEST_DEBUG_PROXY, "Setting importance of %s to %d.\n",
					request->getFile(),
					importance );
		}
        offset = request->getStart();
        filesize = request->getEnd();
        reply = storage_manager->open( &tmp,
									   client->getUser(),
									   request->getFile(),
									   WRITE_FLAGS,
									   offset,
									   filesize,
									   client->getCwd(),
									   link_count,
									   importance );
		if ( reply == NEST_SUCCESS ) {
			*storage = tmp;
		} else {
			*storage = NULL;
		}
        break;
    case CLIENT_RECV_FILE:
    case THIRD_PARTY_PUT:
        offset = request->getStart();
        filesize = -1;
        reply = storage_manager->open( &tmp,
									   client->getUser(),
									   request->getFile(),
									   READ_FLAGS,
									   offset,
									   filesize,
									   client->getCwd(),
									   link_count,
									   importance );
		if ( reply == NEST_SUCCESS ) {
			*storage = tmp;
		} else {
			*storage = NULL;
		}
        break;
    default:
        *storage = NULL;
		nest_debug( NEST_DEBUG_ALWAYS,
					"Unknown stoarge request %d\n", request->getType() );
        break;
    }

    return reply;
}

    // this function is called by both the home server and proxy servers
    // so a pointer is passed to the proxyInit function
    //
    // when we run in proxy mode, we translate client requests into
    // requests for chunks.  however, we also will use transfer file
    // to send the proxy chunk once it arrives from the server
    // so we only want to translate in the first case, but not the second
int
TransferFile( nestClient *client,
			  nestRequest *request,
			  bool already_translated )
{

    nestRequest *req = NULL;
	nestRequest *client_req = request;

	// Handle proxy requests, too...
    if ( serverIsProxy() ) {
        if ( ! already_translated ) {
            proxyTranslateRequest( client );
        }
        req = client->getProxyRequest();
    } else {
        req = request;
    }

    NestReplyStatus status = NEST_SUCCESS;

    int  type         = req->getType();
    nestFileSize end  = req->getEnd();
    const char *file  = req->getFile();

    nest_debug( NEST_DEBUG_TRANSFER,
				"TransferFile %s for %s (%s) (len %d), type %d\n",
				file, client->getUserIdentity(),
				client->getHostIp(), end, type );

    switch( type ) {
    case CLIENT_SEND_FILE:
            // try to reserve the necessary space
            // status = MakeReservation( client );
        break;

    case CLIENT_RECV_FILE:
    case THIRD_PARTY_PUT:
        nestFileSize filelength;
            // get the filelength, if file not found, report this and return
        status = storage_manager->filelength( &filelength,
											  file,
											  client->getUser(),
											  client->getCwd() );
		if ( status != NEST_SUCCESS ) {
            break;
        }

            // check if user is requesting to read the entire file or to read
            // past eof, adjust that here
            // don't forget to adjust in both proxy and normal reqs
        if ( end == ENTIRE_FILE ) {
            end = filelength;
            req->setEnd( filelength );
            client_req->setEnd( filelength );
        } else if ( end > filelength ) {
            nest_debug( NEST_DEBUG_PROXY, 
                        "Req end %ld > filelength %ld. Fixing\n",
                        end, filelength );
            end = filelength;
            //long diff = req->getEnd() - filelength;
                // these next two lines are what proxys should do...
            //req->setEnd( req->getEnd() - diff );
            //request->setEnd(request->getEnd() - diff);
			nestFileSize new_end = filelength + req->getChunk();
			req->setEnd( new_end );
            client_req->setEnd( new_end );
        }
            // special case, user trying to read past eof
        if ( req->getStart() >= filelength + req->getChunk() ) {
                // make it a 0 byte transfer so they will go away
			req->setStart( req->getEnd() );
			client_req->setStart( client_req->getEnd() );
        }
        break;

    case THIRD_PARTY_GET:
            // nothing to do, except when we find out the size, maybe
            // we need to try to make a reservation . . .
        break;

    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
            // nothing to do here
        break;

    default:
        status = NEST_UNKNOWN_REQUEST_TYPE;
        break;
    }

        // deal with proxy logic here to retrieve files we don't have
        // probably should never see this NEST_FLOCK error
    if ( serverIsProxy() &&
		 ( status==NEST_FILE_NOT_FOUND || status==NEST_FLOCK )  ){
        if ( status == NEST_FLOCK ) {
            nest_warn( __FILE__, __FUNCTION__, 
                    "Possibly mishandling an unexpected error at %s,%s: %s",
                    __FILE__, __FUNCTION__, NestErrorString( status ) );
        }
        assert( type == CLIENT_RECV_FILE ); 
        nest_debug( NEST_DEBUG_PROXY, "Will proxy request %s for %s.\n",
                file, client->getUserIdentity() );
        status = addProxyTransfer( client );
        if ( status == NEST_SUCCESS ) {
            return ACTIVE_CONNECTION;
        } else {
            nest_debug( NEST_DEBUG_PROXY, "Failed to proxy request %s: %s.\n",
                    file, NestErrorString( status ) ); 
                // fall through to sendErrorMessage
        }
    }

        // make sure parameters are OK
        // but don't overwrite a previously set error value
	nestFileSize	start = req->getStart( );
    if ( status == NEST_SUCCESS ) {
        if ( (start < 0) || ((end != ENTIRE_FILE) && (end < start)) ) {
            nest_warn( __FILE__,__FUNCTION__,
					   "bad params, %ld end, %ld start\n",
					   end, start );
            status = NEST_BAD_PARAMETERS;
        }
    }
    
        // if no errors yet, add transfer request into concurrency manager
    if ( status == NEST_SUCCESS ) {
            // concurrencyManager now responsible for client->endRequest
        nestProtocol *storage;
        status = openStorageConnection( &storage, client, client_req );

		// Ok?
		if ( status != NEST_SUCCESS ) {
			nest_debug( NEST_DEBUG_TRANSFER,
						"transfer failure: storage error: %s\n",
						NestErrorString( status ) );
		}
		else
		{
			// This is for transfers that don't send filesize info
			if( req->getEnd() == ENTIRE_FILE ) {
				status = storage_manager->increaseAllocation( client );
				if(status != NEST_SUCCESS) {
					storage_manager->cleanupGftpTransfer(client);
				}
			}

			if ( status == NEST_SUCCESS ) {
				client->assignStorageConnection( storage );
				status = concurrency_manager->addTask( client );
			}
		}
    }

    if ( status == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_TRANSFER,
                "transfer added to concurrency manager\n" );
        client->dump( );
        return ACTIVE_CONNECTION;
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        client->endRequest( status );
        nest_debug( NEST_DEBUG_TRANSFER, "transfer failure: %s\n",
                NestErrorString( status ) );
        client->dump( );
        return PASSIVE_CONNECTION;
    }
}

NestReplyStatus
SendAvailableSpace( nestClient *client, nestRequest *request )
{
    nestLotSize avail, total;
    NestReplyStatus status = storage_manager->bytesAvailable(IS_USER,
            &avail, &total );
    if ( status == NEST_SUCCESS ) {
        client->sendDiskspace( avail, total );
    } else {
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }
    return status;
}

int
NestSync( nestClient *client, nestRequest *request )
{
    int connection = PASSIVE_CONNECTION;
    struct timeval before, after;
    nest_debug( NEST_DEBUG_SERVER, "client %s requested a sync\n",
            client->getUserIdentity() );
    NestGettimeofday( &before );
    NestReplyStatus status = storage_manager->sync( client->getUser() );
    NestGettimeofday( &after );

    if( status == NEST_SUCCESS ) {
        nest_debug( NEST_DEBUG_SERVER, "sync successful, took %ld millis\n",
                NestTimediff( &before, &after, __FUNCTION__ ) );
        if ( serverIsProxy() ) {
            status = flushDirtyWrites( client ); 
                // client is now responsibility of proxy server 
            connection = ACTIVE_CONNECTION; 
        } else {
            client->sendAck();
        }
    } else {
        nest_warn( __FILE__, __FUNCTION__,
				   "sync failure: %s\n", NestErrorString( status ) );
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    }

    if ( connection == PASSIVE_CONNECTION ) {
            // otherwise, the proxy code must eventually end the request
        client->endRequest( status );
    }
    return connection;
}


#define PROXY_ITERATOR list<char*>::iterator

// a client (or a proxy monitor) can query a home nest to know which proxy
// nodes have registered themselves for a particular client
NestReplyStatus
QueryProxys( nestClient *client, nestRequest *request )
{
    list< char *> *proxys;

    if ( registered_proxys == NULL ) {
        proxys = NULL;
    } else {
        proxys = (list<char*>*)
            nest_hash_table_lookup( registered_proxys,
									client->getHostIdentity() );
    }

    client->getConnection()->sendProxyList( proxys );

    return NEST_SUCCESS;
    
}

NestReplyStatus
RemoveProxyClient( nestClient *client )
{

    list< char *> *proxys;
    const char *proxy = client->proxyAddress();

    nest_debug( NEST_DEBUG_PROXY, "Removing proxy %s for client %s.\n",
                proxy, client->getHostIdentity() );

        // find the list of proxys registered for this client
    assert( registered_proxys != NULL );
    void *entry = nest_hash_table_lookup( registered_proxys,
										  client->getHostIdentity());
    proxys = (list<char *> *)entry;
    assert( proxys != NULL );

        // remove this particular proxy
    bool found = false;
    PROXY_ITERATOR itr = proxys->begin();
    while ( itr != proxys->end() ) {
        char *g = (char *)*itr;
        if ( ! strcmp( g, proxy ) ) {
            found = true;
            itr = proxys->erase( itr );
            NestFree( g );
            break;
        }
        itr++;
    }
    if ( ! found ) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Tried to remove proxy %s for client %s, not found.\n",
				   proxy, client->getHostIdentity() );
        //assert( found );
    }

    return NEST_SUCCESS;
}

    
// there are two reasons why clients register with a nest:
// 1) nest is home node and client is a proxy server
// 2) nest is proxy coordinator node and client is a proxy server
NestReplyStatus
RegisterProxyClient( nestClient *client, nestRequest *request )
{

    list< char *> *proxys;      // the list of proxys known for a client
    const char *proxy;          // the proxy location being added now

    proxy = request->getProxyLocation();
    client->setProxy( proxy );
    nest_debug(NEST_DEBUG_PROXY,
			   "Client %s (%d) from %s registering proxy %s\n",
			   client->getHostIdentity(), client->getId(),
			   client->getHostIp(), proxy);

        // init the main structure if there already isn't one
    if ( registered_proxys == NULL ) {
        registered_proxys = nest_hash_table_create( 1037, NULL );
    }

        // see if this client has previously registered any proxys
        // if not, create a list for them
    void *entry = nest_hash_table_lookup( registered_proxys,
										  client->getHostIdentity());
    if ( entry == NULL ) {
        proxys = new list< char *>;
        proxys->clear();
        nest_hash_table_insert( registered_proxys, 
                                client->getHostIdentity(), 
                                (void *)proxys );
    } else {
        proxys = (list<char *> *)entry;
    }

        // see whether we need to add this proxy to the list 
    bool found = false;
    for( PROXY_ITERATOR itr = proxys->begin(); itr != proxys->end(); itr++ ) {
        if ( ! strcmp( *itr, proxy ) ) {
            found = true;
            break;
        }
    }
        // important to do a push_back and not a push_front so that
        // new dynamicly larding pfs clients put proxys into same buckets
        // as old pfs dynamically larding clients
    if ( ! found ) proxys->push_back( NestStrdup( proxy ) );

        // if the proxy is coordinator needs to know about this
    if ( serverIsProxy() ) {
            // proxy will deal with sending ack or error reply
        return proxyRegisterProxyClient( client, (char*)proxy );
    }

    //return QueryProxys( client );
    client->sendAck();
    return NEST_SUCCESS;
}

NestReplyStatus
FormatDisk( nestClient *client, nestRequest *request )
{
        // delete all entries and report available space
    NestReplyStatus status = 
        storage_manager->formatDisk( client->getUser() , NULL );
    SendAvailableSpace( client, request );
    return status;
}

NestReplyStatus
SendNestAd( nestClient *client, nestRequest *request )
{

    NestReplyStatus status = NEST_SUCCESS;
        // should we update it first maybe?
    WriteAd();

    nestProtocol *adfile = storage_manager->openMetaFile(AD_NAME,READ_FLAGS);

    if ( adfile == NULL ) {
        status = NEST_REMOTE_FILE_NOT_FOUND; 
        client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
    } else {
        int adlen = storage_manager->metaFilelength( AD_NAME );
        nestAd *ad = advertiser->readAd( adfile, adlen );
        if ( ad == NULL ) {
            status = NEST_REMOTE_FILE_NOT_FOUND; 
            client->sendErrorMessage( __FILE__, __FUNCTION__,  status );
        } else {
            client->sendAd( ad );
            delete( ad );
        }
        delete( adfile );
    }
    return status; 
}

NestReplyStatus
SendDirectoryList( nestClient *client, nestRequest *request )
{

    NestFileInfo *list =
        storage_manager->directoryContents( client->getUser(),
											client->getCwd() );

    client->sendDirectoryList( list );
    while( list != NULL ) {
        NestFileInfo *tmp = list;
        list = list->next;
        NestFree( tmp );
    }
    return NEST_SUCCESS;
}

NestReplyStatus
SendBandwidthStatistics( nestClient *client, nestRequest *request )
{
    client->sendBandwidthStatistics( );
    return NEST_SUCCESS;
}

const char *getRootDir( void )
{
	return storage_manager->getRootDir( );
}
