/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_worker_pool.h"
#include "nest_concurrency_manager.h"
#include "nest_filetransfer.h"

// NOTE that nest_worker and nest_worker_pool object speak nest_protocol
// between them, regardless of what protocol is spoken by the client


nestWorker::nestWorker( pid_t pid,
						int stream,
						int id,
						nestProtocolManager *pm
						)
{
    this->pid    = pid;
    this->id     = id;
    this->idle   = WORKER_IDLE;
    this->stream  = stream;
	this->protocol_manager = pm;
}

pid_t
nestWorker::getPid()
{
    return pid;
} 

bool
nestWorker::isIdle()
{
    return( idle == WORKER_IDLE );
}

void
nestWorker::setIdle( int state )
{
    idle = state;
}

bool
nestWorker::sendMessage( const char * msg )
{
    nest_debug( NEST_DEBUG_POP, "Gonna send %s to worker %d\n", msg, id );
    return SendNestMessage( stream, msg );
}

bool
nestWorker::canReceiveMessage( fd_set *readyfds )
{
    if ( NestFd_isset( stream, readyfds ) ) {
        NestFd_clear( stream, readyfds ); // so we don't loop
        return true;
    } else {
        return false;
    }
}

int 
nestWorker::getID( )
{
    return id;
}

int
nestWorker::getStream()
{
    return stream;
}

NestReplyStatus
nestWorker::assignTask( nestClient *client )
{
    NestReplyStatus status = NEST_SUCCESS;

    this->client         = client;

    if ( client->getConnection()->getType() == GFTP_PROTOCOL
            || client->getConnection()->getType() == FTP_PROTOCOL ) {
        return NEST_NOT_YET_IMPLEMENTED;
    }

    char msg[NEST_MESSAGESIZE];

        // TODO:  Instead of sending weird stuff here, we should have a
        // function that sends a client and a function that receives one
        // we'll need more client info to pass authentication for 3pt's
    int type = client->getRequest()->getType();
    switch( type ) {
    case PROXY_PUT:
    case PROXY_GET:
        nest_debug( NEST_DEBUG_PROXY, "Assigning a proxy task to worker\n" );
            // fall through
    case CLIENT_SEND_FILE:
    case CLIENT_RECV_FILE:
        NestSnprintf( msg, NEST_MESSAGESIZE, "%d %d %d %d %d %s %ld %ld\n",
            type, client->getConnection()->getType(),
            client->getStorageConnection()->getType(),
            client->getId(), client->getConnection()->getSocket(),
            client->getRequest()->getNestFile(),
            client->getRequest()->getStart(),
            client->getRequest()->getLength() );
        break;
    case THIRD_PARTY_PUT:
    case THIRD_PARTY_GET:
        NestSnprintf( msg, NEST_MESSAGESIZE, "%d %d %d %d %d %s %s %s %ld\n",
                type, client->getConnection()->getType(),
                client->getStorageConnection()->getType(),
                client->getId(), client->getConnection()->getSocket(),
                client->getRequest()->getNestFile(),
                client->getRequest()->getRemoteFile(),
                client->getRequest()->getRemoteHost(),
                client->getRequest()->getLength() );
        break;
    case TEST_SERVER_TO_CLIENT:
    case TEST_CLIENT_TO_SERVER:
        NestSnprintf( msg, NEST_MESSAGESIZE, "%d %d %d %d %ld\n",
            type, client->getConnection()->getType(),
            client->getId(), client->getConnection()->getSocket(),
            client->getRequest()->getLength() );
        break;
    case COPY_REQUEST:
        NestSnprintf( msg, NEST_MESSAGESIZE,
					  "%d %d %d %d %d %d %s %s %ld %ld\n",
            type, client->getConnection()->getType(),
            client->getStorageConnection()->getType(),
            client->getOtherStorageConnection()->getType(),
            client->getId(), client->getConnection()->getSocket(),
            client->getRequest()->getNestFile(),
            client->getRequest()->getDestFile( ),
            client->getRequest()->getStart(),
            client->getRequest()->getLength() ); 
            break;
    default:
        nest_die( "unknown transfer type: %d\n", type );
    }

        // send the message and then send the necessary protocol connections
    nest_debug( NEST_DEBUG_POP, "Sending msg %s to w%d\n", msg, getID() );
    sendMessage( msg );
    if ( type != TEST_SERVER_TO_CLIENT && type != TEST_CLIENT_TO_SERVER ) {
        status = protocol_manager->SendConnection(
            stream, client->getStorageConnection() );
        NEST_REPLY_CHECK(status);
        if(type == COPY_REQUEST){
            status = protocol_manager->SendConnection(
            stream, client->getOtherStorageConnection() );
            NEST_REPLY_CHECK(status);
        }
            
    }
    return protocol_manager->SendConnection(
		stream, client->getConnection( ) );
}

void
nestWorker::endTask( bool success, long bytes_transferred )
{
        // TODO: one time this client pointer was bad in gdb...?
    client->getTransfer()->setBytesTransferred( bytes_transferred );
    //client->endTransfer( success );   // taskDone will do this
    this->client = NULL;
    setIdle( true );
}

nestClient *
nestWorker::getClient( )
{
    return client;
}

